// Eightway: A 2.5D graphics engine based on sprite scaling.
// 2020-01-28 No Time To Play <https://notimetoplay.org/>
// Eightway Engine is open source under the MIT License.

var EightWay = (function() {

"use strict";

// The correct value for diagonals is 0.7, i.e. half the unit
// square's, well, diagonal. The wrong but more intuitive 0.5
// may still be useful for a sort of "isometric" look.
var matrices = {
	n: [	1, 0,
		0, 1],
	ne: [	0.7, -0.7,
		0.7,  0.7],
	e: [	0, -1,
		1,  0],
	se: [	-0.7, -0.7,
		0.7, -0.7],
	s: [	-1,  0,
		 0, -1],
	sw: [	-0.7, 0.7,
		-0.7, -0.7],
	w: [	0,  1,
		-1, 0],
	nw: [	0.7, 0.7,
		-0.7, 0.7]
};

function Camera(x, y, z, f) {
	this.x = x || 0;
	this.y = y || 0;
	this.z = z || 0;
	this.f = f || 200; // Default HTML5 canvas width/height.
	
	this.px = 0;
	this.py = 0;
	this.scale = 0;

	this.project = function (x, y, z) {
		x -= this.x;
		y -= this.y;
		z -= this.z;
		
		this.scale = this.f / z;
		this.px = x * this.scale;
		this.py = y * this.scale;
	}
}

function Sprite(id, x, y, z, scale) {
	this.id = id;
	this.x = x;
	this.y = y;
	this.z = z || 0;
	this.scale = scale || 1;
}

function Scene() {
	this.center = {x: 0, y: 0};
	this.spacing = {u: 1, v: 1};
	this.viewDist = 12;
	this.sprites = [];
	this.visible = [];
	
	this.addSprite = function (id, x, y, z, scale) {
		return this.sprites.push(new Sprite(id, x, y, z, scale));
	};
	
	this.render = function (camera, facing) {
		var mat = matrices[facing];
		var ts = this.spacing;
		this.visible.length = 0;
		
		for (var i = 0; i < this.sprites.length; i++) {
			var sprite = this.sprites[i];
			var x = sprite.x - this.center.x;
			var y = sprite.y - this.center.y;
			
			// Include sprites beyond the viewing distance
			// that might end up within it after rotation.
			if (Math.abs(x) > this.viewDist * 1.4)
				continue;
			else if (Math.abs(y) > this.viewDist * 1.4)
				continue;
			
			var x1 = x * mat[0] + y * mat[1];
			var y1 = x * mat[2] + y * mat[3];
			
			// Check again to account for facing a diagonal.
			if (Math.abs(x1) > this.viewDist)
				continue;
			else if (Math.abs(y1) > this.viewDist)
				continue;
			
			camera.project(x1 * ts.u, sprite.z * ts.v, y1 * ts.u);

			if (camera.scale > 0 && camera.scale < Infinity) {
				this.visible.push(new Sprite(
					sprite.id,
					camera.px, camera.py, camera.scale,
					camera.scale * sprite.scale));
			}
		}
		
		this.visible.sort(function (a, b) {
			if (a.z < b.z)
				return -1;
			else if (a.z > b.z)
				return 1;
			else
				return 0;
		});
		
		return this.visible;
	};
}

return {
	"matrices": matrices,
	"Camera": Camera,
	"Sprite": Sprite,
	"Scene": Scene
};

})();
