# Eightway: A 2.5D graphics engine based on sprite scaling.
# 2019-08-29 No Time To Play <https://notimetoplay.org/>
# Eightway Engine is open source under the MIT License.

from __future__ import division

# The correct value for diagonals is 0.7, i.e. half the unit
# square's, well, diagonal. The wrong but more intuitive 0.5
# may still be useful for a sort of "isometric" look.
matrices = {
	"n": [	1, 0,
		0, 1],
	"ne": [	0.7, -0.7,
		0.7,  0.7],
	"e": [	0, -1,
		1,  0],
	"se": [	-0.7, -0.7,
		0.7, -0.7],
	"s": [	-1,  0,
		 0, -1],
	"sw": [	-0.7, 0.7,
		-0.7, -0.7],
	"w": [	0,  1,
		-1, 0],
	"nw": [	0.7, 0.7,
		-0.7, 0.7]
}

class Camera:
	def __init__(self, x, y, z, f):
		self.x = x
		self.y = y
		self.z = z
		self.f = f
		
		self.px = 0
		self.py = 0
		self.scale = 0
	
	def project(self, x, y, z):
		x -= self.x
		y -= self.y
		z -= self.z
		if z != 0:
			self.scale = self.f / z
			self.px = x * self.scale
			self.py = y * self.scale
		else:
			self.scale = 0
			self.px = 0
			self.py = 0

class Scene:
	def __init__(self):
		self.center = {"x": 0, "y": 0}
		self.viewDist = 12
		self.spacing = 1
		self.sprites = []
	
	def addSprite(self, sprite, x, y, z = 0):
		self.sprites.append({
			"sprite": sprite, "x": x, "y": y, "z": z
		})
	
	def render(self, camera, facing):
		mat = matrices[facing]
		ts = self.spacing
		visible = []
		
		for i in self.sprites:
			x = i["x"] - self.center["x"]
			y = i["y"] - self.center["y"]
			
			# Include sprites beyond the viewing distance
			# that might end up within it after rotation.
			if abs(x) > self.viewDist * 1.4:
				continue
			elif abs(y) > self.viewDist * 1.4:
				continue
			
			x1 = x * mat[0] + y * mat[1]
			y1 = x * mat[2] + y * mat[3]

			# Check again to account for facing a diagonal.
			if abs(x1) > self.viewDist:
				continue
			elif abs(y1) > self.viewDist:
				continue
			
			camera.project(x1 * ts, i["z"] * ts, y1 * ts)

			if camera.scale > 0:
				visible.append((
					i["sprite"],
					camera.px,
					camera.py,
					camera.scale
				))
		
		visible.sort(key=lambda v: v[3])
		
		for i in visible:
			yield i
